@@ -22,27 +22,23 @@ implementations.
 =cut
 
 use strict;
-use Scalar::Util 'refaddr';
-use Params::Util '_INSTANCE',
-                 '_ARRAY';
+use Clone 0.30      ();
+use Scalar::Util    qw{refaddr};
+use Params::Util    qw{_INSTANCE _ARRAY};
+use PPI::Util       ();
 use PPI::Node       ();
-use Clone           ();
-use List::MoreUtils ();
-use overload 'bool' => sub () { 1 },
-             '""'   => 'content',
-             '=='   => '__equals',
-             '!='   => '__nequals',
-             'eq'   => '__eq',
-             'ne'   => '__ne';
+use PPI::Singletons '%_PARENT', '%_POSITION_CACHE';
 
-use vars qw{$VERSION $errstr %_PARENT};
-BEGIN {
-	$VERSION = '1.115';
-	$errstr  = '';
+our $VERSION = '1.283';
 
-	# Master Child -> Parent index
-	%_PARENT = ();
-}
+our $errstr = "";
+
+use overload 'bool' => \&PPI::Util::TRUE;
+use overload '""'   => 'content';
+use overload '=='   => '__equals';
+use overload '!='   => '__nequals';
+use overload 'eq'   => '__eq';
+use overload 'ne'   => '__ne';
 
 
 
@@ -58,7 +54,7 @@ BEGIN {
 Because we treat whitespace and other non-code items as Tokens (in order to
 be able to "round trip" the L<PPI::Document> back to a file) the
 C<significant> method allows us to distinguish between tokens that form a
-part of the code, and tokens that arn't significant, such as whitespace,
+part of the code, and tokens that aren't significant, such as whitespace,
 POD, or the portion of a file after (and including) the C<__END__> token.
 
 Returns true if the Element is significant, or false it not.
@@ -66,7 +62,7 @@ Returns true if the Element is significant, or false it not.
 =cut
 
 ### XS -> PPI/XS.xs:_PPI_Element__significant 0.845+
-sub significant { 1 }
+sub significant() { 1 }
 
 =pod
 
@@ -74,7 +70,7 @@ sub significant { 1 }
 
 The C<class> method is provided as a convenience, and really does nothing
 more than returning C<ref($self)>. However, some people have found that
-they appreciate the lazyness of C<$Foo-E<gt>class eq 'whatever'>, so I
+they appreciate the laziness of C<$Foo-E<gt>class eq 'whatever'>, so I
 have caved to popular demand and included it.
 
 Returns the class of the Element as a string
@@ -121,14 +117,14 @@ Returns the basic code as a string (excluding here-doc content).
 =cut
 
 ### XS -> PPI/XS.xs:_PPI_Element__content 0.900+
-sub content { '' }
+sub content() { '' }
 
 
 
 
 
 #####################################################################
-# Naigation Methods
+# Navigation Methods
 
 =pod
 
@@ -147,10 +143,48 @@ sub parent { $_PARENT{refaddr $_[0]} }
 
 =pod
 
+=head2 descendant_of $element
+
+Answers whether a C<PPI::Element> is contained within another one.
+
+C<PPI::Element>s are considered to be descendants of themselves.
+
+=cut
+
+sub descendant_of {
+	my $cursor = shift;
+	my $parent = shift or return undef;
+	while ( refaddr $cursor != refaddr $parent ) {
+		$cursor = $_PARENT{refaddr $cursor} or return '';
+	}
+	return 1;
+}
+
+=pod
+
+=head2 ancestor_of $element
+
+Answers whether a C<PPI::Element> is contains another one.
+
+C<PPI::Element>s are considered to be ancestors of themselves.
+
+=cut
+
+sub ancestor_of {
+	my $self   = shift;
+	my $cursor = shift or return undef;
+	while ( refaddr $cursor != refaddr $self ) {
+		$cursor = $_PARENT{refaddr $cursor} or return '';
+	}
+	return 1;
+}
+
+=pod
+
 =head2 statement
 
 For a C<PPI::Element> that is contained (at some depth) within a
-L<PPI::Statment>, the C<statement> method will return the first parent
+L<PPI::Statement>, the C<statement> method will return the first parent
 Statement object lexically 'above' the Element.
 
 Returns a L<PPI::Statement> object, which may be the same Element if the
@@ -194,6 +228,8 @@ sub top {
 
 =pod
 
+=head2 document
+
 For an Element that is contained within a L<PPI::Document> object,
 the C<document> method will return the top-level Document for the Element.
 
@@ -219,12 +255,10 @@ immediately after the current one, or false if there is no next sibling.
 
 sub next_sibling {
 	my $self     = shift;
-	my $parent   = $_PARENT{refaddr $self} or return '';
 	my $key      = refaddr $self;
+	my $parent   = $_PARENT{$key} or return '';
 	my $elements = $parent->{children};
-	my $position = List::MoreUtils::firstidx {
-		refaddr $_ == $key
-		} @$elements;
+	my $position = $parent->__position($self);
 	$elements->[$position + 1] || '';
 }
 
@@ -242,12 +276,10 @@ sibling.
 
 sub snext_sibling {
 	my $self     = shift;
-	my $parent   = $_PARENT{refaddr $self} or return '';
 	my $key      = refaddr $self;
+	my $parent   = $_PARENT{$key} or return '';
 	my $elements = $parent->{children};
-	my $position = List::MoreUtils::firstidx {
-		refaddr $_ == $key
-		} @$elements;
+	my $position = $parent->__position($self);
 	while ( defined(my $it = $elements->[++$position]) ) {
 		return $it if $it->significant;
 	}
@@ -267,12 +299,10 @@ C<PPI::Element> object.
 
 sub previous_sibling {
 	my $self     = shift;
-	my $parent   = $_PARENT{refaddr $self} or return '';
 	my $key      = refaddr $self;
+	my $parent   = $_PARENT{$key} or return '';
 	my $elements = $parent->{children};
-	my $position = List::MoreUtils::firstidx {
-		refaddr $_ == $key
-		} @$elements;
+	my $position = $parent->__position($self);
 	$position and $elements->[$position - 1] or '';
 }
 
@@ -290,12 +320,10 @@ sibling.
 
 sub sprevious_sibling {
 	my $self     = shift;
-	my $parent   = $_PARENT{refaddr $self} or return '';
 	my $key      = refaddr $self;
+	my $parent   = $_PARENT{$key} or return '';
 	my $elements = $parent->{children};
-	my $position = List::MoreUtils::firstidx {
-		refaddr $_ == $key
-		} @$elements;
+	my $position = $parent->__position($self);
 	while ( $position-- and defined(my $it = $elements->[$position]) ) {
 		return $it if $it->significant;
 	}
@@ -316,7 +344,7 @@ the same object.
 
 Returns a L<PPI::Token> object, or dies on error (which should be extremely
 rare and only occur if an illegal empty L<PPI::Statement> exists below the
-current Element somewhere.
+current Element somewhere.)
 
 =cut
 
@@ -344,7 +372,7 @@ the itself.
 
 Returns a L<PPI::Token> object, or dies on error (which should be extremely
 rare and only occur if an illegal empty L<PPI::Statement> exists below the
-current Element somewhere.
+current Element somewhere.)
 
 =cut
 
@@ -430,9 +458,32 @@ sub previous_token {
 	}
 }
 
+=head2 presumed_features
 
+Returns a hash that indicates which features appear to be active for the given
+element.
 
+=cut
 
+sub presumed_features {
+	my ($self) = @_;
+
+	my @feature_mods;
+	my $walker = $self;
+	while ($walker) {
+		my $sib_walk = $walker;
+		while ($sib_walk) {
+			push @feature_mods, $sib_walk if $sib_walk->can("feature_mods");
+			$sib_walk = $sib_walk->sprevious_sibling;
+		}
+		$walker = $walker->parent;
+	}
+
+	my %feature_mods = map %{$_}, reverse grep defined, map $_->feature_mods,
+	  @feature_mods;
+
+	return \%feature_mods;
+}
 
 #####################################################################
 # Manipulation
@@ -469,39 +520,7 @@ In future, this method may be enhanced to allow the insertion of multiple
 Elements, inline-parsed code strings or L<PPI::Document::Fragment> objects.
 
 Returns true if the Element was inserted, false if it can not be inserted,
-or C<undef> if you do not provide a L<PPI::Element> object as a parameter.
-
-=begin testing __insert_before 6
-
-my $Document = PPI::Document->new( \"print 'Hello World';" );
-isa_ok( $Document, 'PPI::Document' );
-my $semi = $Document->find_first('Token::Structure');
-isa_ok( $semi, 'PPI::Token::Structure' );
-is( $semi->content, ';', 'Got expected token' );
-my $foo = PPI::Token::Word->new('foo');
-isa_ok( $foo, 'PPI::Token::Word' );
-is( $foo->content, 'foo', 'Created Word token' );
-$semi->__insert_before( $foo );
-is( $Document->serialize, "print 'Hello World'foo;",
-	'__insert_before actually inserts' );
-
-=end testing
-
-=begin testing insert_before after __insert_before 6
-
-my $Document = PPI::Document->new( \"print 'Hello World';" );
-isa_ok( $Document, 'PPI::Document' );
-my $semi = $Document->find_first('Token::Structure');
-isa_ok( $semi, 'PPI::Token::Structure' );
-is( $semi->content, ';', 'Got expected token' );
-my $foo = PPI::Token::Word->new('foo');
-isa_ok( $foo, 'PPI::Token::Word' );
-is( $foo->content, 'foo', 'Created Word token' );
-$semi->insert_before( $foo );
-is( $Document->serialize, "print 'Hello World'foo;",
-	'insert_before actually inserts' );
-
-=end testing
+or C<undef> if you do not provide a C<PPI::Element> object as a parameter.
 
 =cut
 
@@ -526,39 +545,7 @@ In future, this method may be enhanced to allow the insertion of multiple
 Elements, inline-parsed code strings or L<PPI::Document::Fragment> objects.
 
 Returns true if the Element was inserted, false if it can not be inserted,
-or C<undef> if you do not provide a L<PPI::Element> object as a parameter.
-
-=begin testing __insert_after 6
-
-my $Document = PPI::Document->new( \"print 'Hello World';" );
-isa_ok( $Document, 'PPI::Document' );
-my $string = $Document->find_first('Token::Quote');
-isa_ok( $string, 'PPI::Token::Quote' );
-is( $string->content, "'Hello World'", 'Got expected token' );
-my $foo = PPI::Token::Word->new('foo');
-isa_ok( $foo, 'PPI::Token::Word' );
-is( $foo->content, 'foo', 'Created Word token' );
-$string->__insert_after( $foo );
-is( $Document->serialize, "print 'Hello World'foo;",
-	'__insert_after actually inserts' );
-
-=end testing
-
-=begin testing insert_after after __insert_after 6
-
-my $Document = PPI::Document->new( \"print 'Hello World';" );
-isa_ok( $Document, 'PPI::Document' );
-my $string = $Document->find_first('Token::Quote');
-isa_ok( $string, 'PPI::Token::Quote' );
-is( $string->content, "'Hello World'", 'Got expected token' );
-my $foo = PPI::Token::Word->new('foo');
-isa_ok( $foo, 'PPI::Token::Word' );
-is( $foo->content, 'foo', 'Created Word token' );
-$string->insert_after( $foo );
-is( $Document->serialize, "print 'Hello World'foo;",
-	'insert_after actually inserts' );
-
-=end testing
+or C<undef> if you do not provide a C<PPI::Element> object as a parameter.
 
 =cut
 
@@ -589,7 +576,7 @@ sub remove {
 
 =head2 delete
 
-For a given C<PPI::Element>, the C<remove> method will remove it from its
+For a given C<PPI::Element>, the C<delete> method will remove it from its
 parent, immediately deleting the C<Element> and all of its children (if it
 has any).
 
@@ -616,12 +603,14 @@ To prevent accidental damage to code, in this initial implementation the
 replacement element B<must> be of the same class (or a subclass) as the
 one being replaced.
 
+If successful, returns the replace element.  Otherwise, returns C<undef>.
+
 =cut
 
 sub replace {
 	my $self    = ref $_[0] ? shift : return undef;
-	my $Element = _INSTANCE(shift, ref $self) or return undef;
-	die "The ->replace method has not yet been implemented";
+	my $replace = _INSTANCE(shift, ref $self) or return undef;
+	return $self->parent->replace_child( $self, $replace );
 }
 
 =pod
@@ -633,24 +622,145 @@ indexed the Element locations using C<PPI::Document::index_locations>, the
 C<location> method will return the location of the first character of the
 Element within the Document.
 
-Returns the location as a reference to a three-element array in the form
-C<[ $line, $rowchar, $col ]>. The values are in a human format, with the
-first character of the file located at C<[ 1, 1, 1 ]>. 
+Returns the location as a reference to a five-element array in the form C<[
+$line, $rowchar, $col, $logical_line, $logical_file_name ]>. The values are in
+a human format, with the first character of the file located at C<[ 1, 1, 1, ?,
+'something' ]>.
 
 The second and third numbers are similar, except that the second is the
-literal horozontal character, and the third is the visual column, taking
-into account tabbing.
+literal horizontal character, and the third is the visual column, taking
+into account tabbing (see L<PPI::Document/"tab_width [ $width ]">).
 
-Returns C<undef> on error, or if the L<PPI::Document> object has not been indexed.
+The fourth number is the line number, taking into account any C<#line>
+directives.  The fifth element is the name of the file that the element was
+found in, if available, taking into account any C<#line> directives.
+
+Returns C<undef> on error, or if the L<PPI::Document> object has not been
+indexed.
 
 =cut
 
 sub location {
 	my $self = shift;
+
+	$self->_ensure_location_present or return undef;
+
+	# Return a copy, not the original
+	return [ @{$self->{_location}} ];
+}
+
+=pod
+
+=head2 line_number
+
+If the Element exists within a L<PPI::Document> that has indexed the Element
+locations using C<PPI::Document::index_locations>, the C<line_number> method
+will return the line number of the first character of the Element within the
+Document.
+
+Returns C<undef> on error, or if the L<PPI::Document> object has not been
+indexed.
+
+=cut
+
+sub line_number {
+	my $self = shift;
+
+	my $location = $self->location() or return undef;
+	return $location->[0];
+}
+
+=pod
+
+=head2 column_number
+
+If the Element exists within a L<PPI::Document> that has indexed the Element
+locations using C<PPI::Document::index_locations>, the C<column_number> method
+will return the column number of the first character of the Element within the
+Document.
+
+Returns C<undef> on error, or if the L<PPI::Document> object has not been
+indexed.
+
+=cut
+
+sub column_number {
+	my $self = shift;
+
+	my $location = $self->location() or return undef;
+	return $location->[1];
+}
+
+=pod
+
+=head2 visual_column_number
+
+If the Element exists within a L<PPI::Document> that has indexed the Element
+locations using C<PPI::Document::index_locations>, the C<visual_column_number>
+method will return the visual column number of the first character of the
+Element within the Document, according to the value of
+L<PPI::Document/"tab_width [ $width ]">.
+
+Returns C<undef> on error, or if the L<PPI::Document> object has not been
+indexed.
+
+=cut
+
+sub visual_column_number {
+	my $self = shift;
+
+	my $location = $self->location() or return undef;
+	return $location->[2];
+}
+
+=pod
+
+=head2 logical_line_number
+
+If the Element exists within a L<PPI::Document> that has indexed the Element
+locations using C<PPI::Document::index_locations>, the C<logical_line_number>
+method will return the line number of the first character of the Element within
+the Document, taking into account any C<#line> directives.
+
+Returns C<undef> on error, or if the L<PPI::Document> object has not been
+indexed.
+
+=cut
+
+sub logical_line_number {
+	my $self = shift;
+
+	return $self->location()->[3];
+}
+
+=pod
+
+=head2 logical_filename
+
+If the Element exists within a L<PPI::Document> that has indexed the Element
+locations using C<PPI::Document::index_locations>, the C<logical_filename>
+method will return the logical file name containing the first character of the
+Element within the Document, taking into account any C<#line> directives.
+
+Returns C<undef> on error, or if the L<PPI::Document> object has not been
+indexed.
+
+=cut
+
+sub logical_filename {
+	my $self = shift;
+
+	my $location = $self->location() or return undef;
+	return $location->[4];
+}
+
+sub _ensure_location_present {
+	my $self = shift;
+
 	unless ( exists $self->{_location} ) {
 		# Are we inside a normal document?
 		my $Document = $self->document or return undef;
-		unless ( $Document->isa('PPI::Document::Fragment') ) {
+		if ( $Document->isa('PPI::Document::Fragment') ) {
 			# Because they can't be serialized, document fragments
 			# do not support the concept of location.
 			return undef;
@@ -666,8 +776,7 @@ sub location {
 		}
 	}
 
-	# Return a copy, not the original
-	return [ @{$self->{_location}} ];
+	return 1;
 }
 
 # Although flush_locations is only publically a Document-level method,
@@ -691,7 +800,7 @@ sub _flush_locations {
 			return 1 unless $Token->{_location};
 			next unless refaddr($Token) == refaddr($start);
 
-			# Found the start. Flush it's location
+			# Found the start. Flush its location
 			delete $$Token->{_location};
 			last;
 		}
@@ -749,8 +858,10 @@ sub _clear {
 # ->delete means our reference count has probably fallen to zero.
 # Therefore we don't need to remove ourselves from our parent,
 # just the index ( just in case ).
-### XS -> PPI/XS.xs:_PPI_Element__DESTROY 0.900+
-sub DESTROY { delete $_PARENT{refaddr $_[0]} }
+sub DESTROY {
+  delete $_PARENT{refaddr $_[0]};
+  delete $_POSITION_CACHE{refaddr $_[0]};
+}
 
 # Operator overloads
 sub __equals  { ref $_[1] and refaddr($_[0]) == refaddr($_[1]) }
@@ -777,15 +888,15 @@ as error handlers.
 
 =head1 SUPPORT
 
-See the L<support section|PPI/SUPPORT> in the main module
+See the L<support section|PPI/SUPPORT> in the main module.
 
 =head1 AUTHOR
 
-Adam Kennedy, L<http://ali.as/>, cpan@ali.as
+Adam Kennedy E<lt>adamk@cpan.orgE<gt>
 
 =head1 COPYRIGHT
 
-Copyright (c) 2001 - 2005 Adam Kennedy. All rights reserved.
+Copyright 2001 - 2011 Adam Kennedy.
 
 This program is free software; you can redistribute
 it and/or modify it under the same terms as Perl itself.
