@@ -38,11 +38,9 @@ the standard L<PPI::Structure>, L<PPI::Node> and L<PPI::Element> methods.
 use strict;
 use PPI::Structure ();
 
-use vars qw{$VERSION @ISA};
-BEGIN {
-	$VERSION = '1.220';
-	@ISA     = 'PPI::Structure';
-}
+our $VERSION = '1.283';
+
+our @ISA = "PPI::Structure";
 
 1;
 
