@@ -24,7 +24,7 @@ and newlines may not matter in most Perl code, but they matter in here-docs.
 They are also tricky to store as an object. They look sort of like an
 operator and a string, but they don't act like it. And they have a second
 section that should be something like a separate token, but isn't because a
-strong can span from above the here-doc content to below it.
+string can span from above the here-doc content to below it.
 
 So when parsing, this is what we do.
 
@@ -72,7 +72,7 @@ but for the most part they should more or less DWYM.
 Although technically it can be considered a quote, for the time being
 C<HereDocs> are being treated as a completely separate C<Token> subclass,
 and will not be found in a search for L<PPI::Token::Quote> or
-L<PPI::Token::QuoteLike objects>.
+L<PPI::Token::QuoteLike> objects.
 
 This may change in the future, with it most likely to end up under
 QuoteLike.
@@ -85,12 +85,10 @@ have a relatively large number of unique methods all of their own.
 =cut
 
 use strict;
-use base 'PPI::Token';
 
-use vars qw{$VERSION};
-BEGIN {
-	$VERSION = '1.203';
-}
+our $VERSION = '1.283';
+
+our @ISA = "PPI::Token";
 
 
 
@@ -112,11 +110,20 @@ the here-doc, B<excluding> the terminator line.
 
 =cut
 
-sub heredoc {
-	wantarray
-		? @{shift->{_heredoc}}
-		: scalar @{shift->{_heredoc}};
-}
+sub heredoc { @{shift->{_heredoc}} }
+
+=pod
+
+=head2 indentation
+
+The C<indentation> method returns the indentation string of an indented
+here-doc if that can be determined. If the indented here-doc is damaged
+(say, missing terminator) or the here-doc was not indented, it returns
+C<undef>.
+
+=cut
+
+sub indentation { shift->{_indentation} }
 
 =pod
 
@@ -134,6 +141,25 @@ sub terminator {
 	shift->{_terminator};
 }
 
+sub _is_terminator {
+	my ( $self, $terminator, $line, $indented ) = @_;
+	if ( $indented ) {
+		return $line =~ /^\s*\Q$terminator\E$/;
+	} else {
+		return $line eq $terminator;
+	}
+}
+
+sub _indent {
+	my ( $self, $token ) = @_;
+	my ($indent) = $token->{_terminator_line} =~ /^(\s*)/;
+	return $indent;
+}
+
+sub _is_match_indent {
+	my ( $self, $token, $indent ) = @_;
+	return (grep { /^$indent/ || $_ eq "\n" } @{$token->{_heredoc}}) == @{$token->{_heredoc}};
+}
 
 
 
@@ -143,7 +169,7 @@ sub terminator {
 
 # Parse in the entire here-doc in one call
 sub __TOKENIZER__on_char {
-	my $t     = $_[1];
+	my ( $self, $t ) = @_;
 
 	# We are currently located on the first char after the <<
 
@@ -151,8 +177,9 @@ sub __TOKENIZER__on_char {
 	### FIXME - This regex, and this method in general, do not yet allow
 	### for the null here-doc, which terminates at the first
 	### empty line.
-	my $rest_of_line = substr( $t->{line}, $t->{line_cursor} );
-	unless ( $rest_of_line =~ /^( \s* (?: "[^"]*" | '[^']*' | `[^`]*` | \\?\w+ ) )/x  ) {
+	pos $t->{line} = $t->{line_cursor};
+
+	if ( $t->{line} !~ m/\G( ~? \s* (?: "[^"]*" | '[^']*' | `[^`]*` | \\?\w+ ) )/gcx ) {
 		# Degenerate to a left-shift operation
 		$t->{token}->set_class('Operator');
 		return $t->_finalize_token->__TOKENIZER__on_char( $t );
@@ -167,80 +194,105 @@ sub __TOKENIZER__on_char {
 	# Find the terminator, clean it up and determine
 	# the type of here-doc we are dealing with.
 	my $content = $token->{content};
-	if ( $content =~ /^\<\<(\w+)$/ ) {
+	if ( $content =~ /^\<\<(~?)(\w+)$/ ) {
 		# Bareword
 		$token->{_mode}       = 'interpolate';
-		$token->{_terminator} = $1;
+		$token->{_indented}   = 1 if $1 eq '~';
+		$token->{_terminator} = $2;
 
-	} elsif ( $content =~ /^\<\<\s*\'(.*)\'$/ ) {
+	} elsif ( $content =~ /^\<\<(~?)\s*\'(.*)\'$/ ) {
 		# ''-quoted literal
 		$token->{_mode}       = 'literal';
-		$token->{_terminator} = $1;
+		$token->{_indented}   = 1 if $1 eq '~';
+		$token->{_terminator} = $2;
 		$token->{_terminator} =~ s/\\'/'/g;
 
-	} elsif ( $content =~ /^\<\<\s*\"(.*)\"$/ ) {
+	} elsif ( $content =~ /^\<\<(~?)\s*\"(.*)\"$/ ) {
 		# ""-quoted literal
 		$token->{_mode}       = 'interpolate';
-		$token->{_terminator} = $1;
+		$token->{_indented}   = 1 if $1 eq '~';
+		$token->{_terminator} = $2;
 		$token->{_terminator} =~ s/\\"/"/g;
 
-	} elsif ( $content =~ /^\<\<\s*\`(.*)\`$/ ) {
+	} elsif ( $content =~ /^\<\<(~?)\s*\`(.*)\`$/ ) {
 		# ``-quoted command
 		$token->{_mode}       = 'command';
-		$token->{_terminator} = $1;
+		$token->{_indented}   = 1 if $1 eq '~';
+		$token->{_terminator} = $2;
 		$token->{_terminator} =~ s/\\`/`/g;
 
-	} elsif ( $content =~ /^\<\<\\(\w+)$/ ) {
+	} elsif ( $content =~ /^\<\<(~?)\\(\w+)$/ ) {
 		# Legacy forward-slashed bareword
 		$token->{_mode}       = 'literal';
-		$token->{_terminator} = $1;
+		$token->{_indented}   = 1 if $1 eq '~';
+		$token->{_terminator} = $2;
 
 	} else {
 		# WTF?
 		return undef;
 	}
 
-	# Define $line outside of the loop, so that if we encounter the
-	# end of the file, we have access to the last line still.
-	my $line;
-
 	# Suck in the HEREDOC
-	$token->{_heredoc} = [];
+	$token->{_heredoc} = \my @heredoc;
 	my $terminator = $token->{_terminator} . "\n";
-	while ( defined($line = $t->_get_line) ) {
-		if ( $line eq $terminator ) {
+	while ( defined( my $line = $t->_get_line ) ) {
+		if ( $self->_is_terminator( $terminator, $line, $token->{_indented} ) ) {
 			# Keep the actual termination line for consistency
 			# when we are re-assembling the file
 			$token->{_terminator_line} = $line;
 
+			if ( $token->{_indented} ) {
+				my $indent = $self->_indent( $token );
+				$token->{_indentation} = $indent;
+				# Indentation of here-doc doesn't match delimiter
+				unless ( $self->_is_match_indent( $token, $indent ) ) {
+					push @heredoc, $line;
+					last;
+				}
+
+				s/^$indent// for @heredoc, $token->{_terminator_line};
+			}
+
 			# The HereDoc is now fully parsed
 			return $t->_finalize_token->__TOKENIZER__on_char( $t );
 		}
 
 		# Add the line
-		push @{$token->{_heredoc}}, $line;
+		push @heredoc, $line;
 	}
 
 	# End of file.
 	# Error: Didn't reach end of here-doc before end of file.
-	# $line might be undef if we get NO lines.
-	if ( defined $line and $line eq $token->{_terminator} ) {
-		# If the last line matches the terminator
-		# but is missing the newline, we want to allow
-		# it anyway (like perl itself does). In this case
-		# perl would normally throw a warning, but we will
-		# also ignore that as well.
-		pop @{$token->{_heredoc}};
-		$token->{_terminator_line} = $line;
-	} else {
-		# The HereDoc was not properly terminated.
-		$token->{_terminator_line} = undef;
 
-		# Trim off the trailing whitespace
-		if ( defined $token->{_heredoc}->[-1] and $t->{source_eof_chop} ) {
-			chop $token->{_heredoc}->[-1];
+	# If the here-doc block is not empty, look at the last line to determine if
+	# the here-doc terminator is missing a newline (which Perl would fail to
+	# compile but is easy to detect) or if the here-doc block was just not
+	# terminated at all (which Perl would fail to compile as well).
+	$token->{_terminator_line} = undef;
+	if ( @heredoc and defined $heredoc[-1] ) {
+		# See PPI::Tokenizer, the algorithm there adds a space at the end of the
+		# document that we need to make sure we remove.
+		if ( $t->{source_eof_chop} ) {
+			chop $heredoc[-1];
 			$t->{source_eof_chop} = '';
 		}
+
+		# Check if the last line of the file matches the terminator without
+		# newline at the end. If so, remove it from the content and set it as
+		# the terminator line.
+		$token->{_terminator_line} = pop @heredoc
+			if $self->_is_terminator( $token->{_terminator}, $heredoc[-1], $token->{_indented} );
+	}
+
+	if ( $token->{_indented} && $token->{_terminator_line} ) {
+		my $indent = $self->_indent( $token );
+		$token->{_indentation} = $indent;
+		if ( $self->_is_match_indent( $token, $indent ) ) {
+			# Remove indent from here-doc as much as possible
+			s/^$indent// for @heredoc;
+		}
+
+		s/^$indent// for $token->{_terminator_line};
 	}
 
 	# Set a hint for PPI::Document->serialize so it can
@@ -275,7 +327,7 @@ Adam Kennedy E<lt>adamk@cpan.orgE<gt>
 
 =head1 COPYRIGHT
 
-Copyright 2001 - 2008 Adam Kennedy.
+Copyright 2001 - 2011 Adam Kennedy.
 
 This program is free software; you can redistribute
 it and/or modify it under the same terms as Perl itself.
